
/**
 * The type of messages our frames our sending
 * @type {String}
 */
const MESSAGE_TYPE = 'application/x-postmate-v1+json';

/**
 * Postmate logging function that enables/disables via config
 * @param  {Object} ...args Rest Arguments
 * @return {undefined}
 */
function log(...args) {
  console.log(...args); // eslint-disable-line no-console
}

/**
 * Takes a context, and searches for a value by the key
 * @param  {Object} context The dictionary to search against
 * @param  {String} key     A path within a dictionary (i.e. 'window.location.href')
 * @return {Promise}
 */
function resolveValue(context, key, data) {
  const unwrappedContext = typeof context[key] === 'function' ? context[key](data) : context[key];
  return Promise.resolve(unwrappedContext);
}

/**
 * Composes an API namespace for the respective client or host
 * @param {Object} info Information on the consumer
 */
function Consumer(info) {
  const { id, context, client, frame, host } = info;
  // this._messageID = 1;

  // var target = consumer === 'client' ? host : client;
  log('Registering consumer', id);
  if (id === 'client') {
    log('Client awaiting messages...');
    return {
      frame,
      get(key, data) {
        return new Promise(resolve => {
          // Extract data from response and kill listeners
          const transact = e => {
            if (e.data.key === key) {
              client.removeEventListener('message', transact, false);
              resolve(e.data.value);
            }
          };

          // Prepare for response from Host...
          client.addEventListener('message', transact, false);

          // Then ask host for information
          host.postMessage({
            data,
            key,
            type: MESSAGE_TYPE,
          }, host.location.origin);
        });
      },
    };
  } else if (id === 'host') {
    log('Host awaiting messages...');

    host.addEventListener('message', e => {
      if (e.origin !== client.location.origin) return;
      log('Host received message', e.data);

      const { key, data } = e.data;

      // Reply to Client
      resolveValue(context, key, data)
        .then(value => e.source.postMessage({
          key,
          value,
          type: MESSAGE_TYPE,
        }, e.origin));
    });

    return {};
  }
}

/**
 * The entry point of the Host. Exposes a context, or set of properties that the
 * Client may interact with.
 * @param {Object} context Hash of values, functions, or promises
 */
function Postmate(context) {
  this.id = 'host';
  this.host = window;
  this.context = context;
  this.client = this.host.parent;

  return new Promise((resolve, reject) => {
    const shake = e => {
      if (e.data.postmate === 'handshake') {
        log('Host: Received handshake from Client');
        this.host.removeEventListener('message', shake, false);
        log('Host: Sending handshake reply to Client');
        e.source.postMessage({
          postmate: 'handshake-reply',
          type: MESSAGE_TYPE,
        }, e.origin);
        return resolve(new Consumer(this));
      }
      return reject('Handshake Reply Failed');
    };
    this.host.addEventListener('message', shake, false);
  });
}

/**
 * The entry point of the Client. Begins the handshake strategy
 * @param {Object} userOptions The element to inject the frame into, and the url
 */
Postmate.Handshake = function Handshake(userOptions) {
  const { container, url } = userOptions;

  this.id = 'client';
  this.client = window;
  this.frame = document.createElement('iframe');
  container.appendChild(this.frame);
  this.host = this.frame.contentWindow;

  return new Promise((resolve, reject) => {
    const reply = e => {
      if (e.data.postmate === 'handshake-reply') {
        log('Client: Received handshake reply from Host');
        this.client.removeEventListener('message', reply, false);
        return resolve(new Consumer(this));
      }
      log('Client: Invalid handshake reply');
      return reject('Failed handshake');
    };

    this.client.addEventListener('message', reply, false);

    this.frame.onload = () => {
      log('Client: Sending handshake');
      this.host.postMessage({
        postmate: 'handshake',
        type: MESSAGE_TYPE,
      }, '*');
    };

    log('Client: Loading frame');
    this.frame.src = url;
  });
};

// Export
export default Postmate;
